//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Processed/ReSample.h
//! @brief     Defines class ReSample.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif // SWIG
#ifndef BORNAGAIN_RESAMPLE_PROCESSED_RESAMPLE_H
#define BORNAGAIN_RESAMPLE_PROCESSED_RESAMPLE_H

#include "Base/Type/OwningVector.h"
#include "Resample/Slice/SliceStack.h"
#include <heinz/Vectors3D.h>

class Fluxes;
class Material;
class MultiLayer;
class ReLayout;
class SimulationOptions;

//! Data structure that contains all the necessary data for scattering calculations.
//!
//! If the usage of average materials is requested, layers and particles are sliced into multiple
//! slices and the average material is calculated for each slice.

class ReSample {
public:
    //! Factory method that wraps the private constructor.
    static ReSample make(const MultiLayer& sample, const SimulationOptions& options,
                         bool forcePolarized = false);
    static ReSample make(const MultiLayer& sample); // used by tests only?

    ReSample(const ReSample&) = delete;
    ~ReSample();

    size_t numberOfSlices() const;
    const SliceStack& averageSlices() const;
    const Slice& avgeSlice(size_t i) const;
    const OwningVector<const ReLayout>& relayouts() const;
    double sliceTopZ(size_t i) const;
    double sliceBottomZ(size_t i) const;
    bool polarizing() const; //!< Contains magnetic material, or nonzero magnetic field
    bool hasRoughness() const;
    const MultiLayer& sample() const { return m_sample; }

    Fluxes fluxesIn(const R3& k) const;
    Fluxes fluxesOut(const R3& k) const;

    //! Fourier transform of the correlation function of roughnesses between the interfaces
    double crossCorrSpectralFun(R3 kvec, size_t j, size_t k) const;

private:
    ReSample(const MultiLayer& sample, bool polarized, OwningVector<const ReLayout>&& layouts,
             SliceStack refined_stack);
    const MultiLayer& m_sample;
    const bool m_polarized;
    const OwningVector<const ReLayout> m_relayouts;
    const SliceStack m_stack;
};

#endif // BORNAGAIN_RESAMPLE_PROCESSED_RESAMPLE_H
